const { isEqual } = require('lodash');

const { registerDecider } = require('../validation-pools');
const { registerMatrixEventHandler, sendMatrixText, sendMatrixEvent } = require('../../matrix-bot');
const { matrixPools, matrixUserToAuthorAddress } = require('../../util/db');
const {
  rollup, wallet,
} = require('../../util/contracts');
const read = require('../../util/forum/read');
const { availabilityStakeDuration } = require('./config');
const {
  stakeRollupAvailability, authorsMatch, validatePost,
} = require('./utils');
const computeMatrixPoolResult = require('./matrix-pools/compute-result');
const { initializeBatchItems, addBatchItem, clearBatchItems } = require('./batch/batch-items');
const { getCurrentBatchWorker, initializeBatchWorker } = require('./batch/batch-worker');
const initiateMatrixPool = require('./matrix-pools/initiate');
const { initiateMatrixPools } = require('./matrix-pools/initiate-matrix-pools');
const computeAuthorWeights = require('./batch/compute-author-weights');

const start = async () => {
  console.log('registering validation pool decider for rollup');
  registerDecider(async (pool, post) => {
    // If this is not sent by the work1 contract, it's not of interest here.
    if (pool.sender !== rollup.target) return false;

    // A rollup post should contain
    // - a list of off-chain validation pools
    // - authorship corresponding to the result of those off-chain pools
    if (!post.embeddedData?.batchItems) return false;

    // Our task here is to check whether the posted result agrees with our own computations
    let expectedAuthors;
    try {
      expectedAuthors = await computeAuthorWeights(post.embeddedData.batchItems);
    } catch (e) {
      console.error('Error calculating batch post author weights', e);
      return null;
    }
    const valid = authorsMatch(post.authors, expectedAuthors);
    console.log(`batch post ${pool.props.postId} is ${valid ? 'valid' : 'invalid'}`);
    return valid;
  });

  // Even if we're not the current batch worker, keep track of batch items
  initializeBatchItems();

  // Check for an assigned batch worker
  await initializeBatchWorker();

  // Stake availability and set an interval to maintain it
  await stakeRollupAvailability();
  setInterval(stakeRollupAvailability, availabilityStakeDuration * 1000);

  // Initiate any matrix pools that haven't already occurred
  await initiateMatrixPools();

  /// `sender` is the address that called Rollup.addItem on chain, i.e. the Work2 contract.
  rollup.on('BatchItemAdded', async (postId, sender, fee) => {
    // If we are the batch worker or there is no batch worker, initiate a matrix pool
    const batchWorker = getCurrentBatchWorker();
    if (batchWorker === await wallet.getAddress()
      || batchWorker === '0x0000000000000000000000000000000000000000') {
      let post;
      try {
        post = await read(postId);
      } catch (e) {
        console.error(`Post ID ${postId} not found`);
        return;
      }

      // Initialize a matrix pool
      try {
        await matrixPools.get(postId);
        // If this doesn't throw, it means we or someone else already sent this event
        console.log(`Matrix pool start event has already been sent for postId ${postId}`);
      } catch (e) {
        if (e.status === 404) {
          await initiateMatrixPool(postId, post, sender, fee);
        } else {
          throw e;
        }
      }
    }
  });

  registerMatrixEventHandler(async (client, roomId, event) => {
    switch (event.type) {
      case 'io.dgov.pool.start': {
        // Note that matrix pools are identified by the postId to which they pertain.
        // This means that for a given post there can only be one matrix pool at a time.
        const { postId, sender, ...params } = event.content;
        // We can use LevelDB to store information about validation pools
        const eventId = event.event_id;
        console.log('Matrix pool started', { postId, ...params });
        // Validate the target post, and stake for/against
        let post;
        try {
          post = await read(postId);
        } catch (e) {
          console.error(`Post ID ${postId} not found`);
          break;
        }
        // Register our own stake and send a message
        const { amount, inFavor } = await validatePost(sender, post);
        sendMatrixEvent('io.dgov.pool.stake', { postId, amount, inFavor });
        const matrixPool = {
          postId,
          roomId,
          eventId,
          ...params,
          stakes: [{ amount, inFavor, account: await wallet.getAddress() }],
        };
        await matrixPools.put(postId, matrixPool);
        break;
      }
      case 'io.dgov.pool.stake': {
        const { postId, amount, inFavor } = event.content;
        let account;
        try {
          account = await matrixUserToAuthorAddress(event.sender);
        } catch (e) {
          // Error, sender has not registered their matrix identity
          sendMatrixText(`Matrix user ${event.sender} has not registered their wallet address`);
          break;
        }
        let matrixPool;
        try {
          matrixPool = await matrixPools.get(postId);
        } catch (e) {
          // Error. matrix pool not found
          sendMatrixText(`Received stake for unknown matrix pool, for post ${postId}. Stake sent by ${event.sender}`);
          break;
        }
        const stake = { account, amount, inFavor };
        matrixPool.stakes = matrixPool.stakes ?? [];
        matrixPool.stakes.push(stake);
        await matrixPools.put(postId, matrixPool);
        console.log(`registered stake in matrix pool for post ${postId} by ${account}`);
        break;
      }
      case 'io.dgov.pool.result': {
        // This should be sent by the current batch worker
        // const { stakedFor, stakedAgainst, totalSupply, votePasses, quorumMet, } = result;
        const { postId, result } = event.content;
        let matrixPool;
        try {
          matrixPool = await matrixPools.get(postId);
        } catch (e) {
          // Error. matrix pool not found
          sendMatrixText(`Received result for unknown matrix pool, for post ${postId}. Result sent by ${event.sender}`);
          break;
        }
        // Compare batch worker's result with ours to verify and provide early warning
        const expectedResult = await computeMatrixPoolResult(matrixPool);
        if (!isEqual(result, expectedResult)) {
          sendMatrixText(`Unexpected result for matrix pool, for post ${postId}. Result sent by ${event.sender}\n\n`
            + `received ${JSON.stringify(result)}\n`
            + `expected ${JSON.stringify(expectedResult)}`);
        }
        matrixPool.result = result;
        await matrixPools.put(postId, matrixPool);
        await addBatchItem(postId);
        break;
      }
      case 'io.dgov.rollup.submit': {
        // This should include the identifier of the on-chain validation pool
        const { batchPostId, batchItems, authors } = event.content;
        // Compare batch worker's result with ours to verify
        const expectedAuthors = await computeAuthorWeights(batchItems);
        if (!authorsMatch(authors, expectedAuthors)) {
          sendMatrixText(`Unexpected result for batch post ${batchPostId}`);
        }
        // Reset batchItems in preparation for next batch
        await clearBatchItems(batchItems);
        break;
      }
      default:
    }
  });
};

module.exports = {
  start,
};
