const Promise = require('bluebird');

const { dao, wallet } = require('../util/contracts');
const read = require('../util/forum/read');
const gateByProposal = require('../util/gate-by-proposal');

const {
  ENABLE_STAKING,
} = process.env;

const deciders = [];

const registerDecider = (decider) => {
  deciders.push(decider);
};

let enableStaking;

if (ENABLE_STAKING === 'false') {
  console.log('STAKING DISABLED');
  enableStaking = false;
} else {
  gateByProposal(((enable) => {
    enableStaking = enable;
  }));
}

const start = async () => {
  dao.on('ValidationPoolInitiated', async (poolIndex) => {
    console.log('Validation Pool Initiated, index', poolIndex);
    const pool = await dao.validationPools(poolIndex);
    // Read post from database
    let post;
    try {
      post = await read(pool.props.postId);
    } catch (e) {
      // Post not found
      console.error(`error: post for validation pool ${poolIndex} not found`);
      return;
    }
    console.log('postId:', pool.props.postId);
    console.log('post.content:', post.content);
    // We have the opportunity to stake for/against this validation pool.
    // To implement the legislative process of upgrading this protocol,
    // the execution of this code can be protected by a given proposal.
    // The code will only execute if the proposal has been accepted.
    if (!enableStaking) {
      return;
    }
    const decisions = await Promise.mapSeries(deciders, (decider) => decider(pool, post));
    const inFavor = decisions.some((x) => x === true);
    const nullResult = decisions.some((x) => x === null);
    const currentRep = await dao.balanceOf(await wallet.getAddress());
    let stakeAmount = currentRep;
    if (!inFavor && nullResult) {
      console.log(`Obtained a NULL RESULT for pool ${poolIndex}.`);
      // TODO: Retry?
      // TODO: Notify
      // Calculate the minimum stake S against the post, such that if the honest actors
      // each stake S, the result will be enough to meet the win ratio.
      // This way, we combat the threat of a truly invalid post,
      // while reducing our exposure in the case that the error is unique to us.
      // Assume 2/3 honest actors.
      // S * (2/3) = 1/3
      // S = 1/2;
      stakeAmount = Math.ceil(currentRep / 2);
    }
    // Stake all available reputation
    console.log(`STAKING ${stakeAmount} ${inFavor ? 'in favor of' : 'against'} pool ${poolIndex}`);
    try {
      await dao.stakeOnValidationPool(poolIndex, stakeAmount, inFavor);
    } catch (e) {
      // Maybe the end time passed?
      console.error(`STAKING failed, reason: ${e.reason}`);
    }
  });
};

module.exports = {
  start,
  registerDecider,
};
