// SPDX-License-Identifier: Unlicense
pragma solidity ^0.8.24;

import "./core/DAO.sol";
import "./core/Forum.sol";
import "./Availability.sol";
import "./Proposals.sol";
import "./interfaces/IOnProposalAccepted.sol";

abstract contract Work is Availability, IOnProposalAccepted {
    enum WorkStatus {
        Requested,
        EvidenceSubmitted,
        ApprovalSubmitted,
        Complete
    }

    struct WorkRequest {
        address customer;
        uint256 fee;
        WorkStatus status;
        uint stakeIndex;
        string requestPostId;
        string evidencePostId;
        Citation[] citations;
        bool approval;
    }

    struct PriceProposal {
        uint price;
        uint proposalIndex;
    }

    Proposals proposalsContract;
    uint public price;
    mapping(uint => PriceProposal) public priceProposals;
    uint public priceProposalCount;
    mapping(uint => WorkRequest) public requests;
    uint public requestCount;

    uint constant POOL_DURATION = 20;

    event WorkAssigned(uint requestIndex, uint stakeIndex);
    event WorkEvidenceSubmitted(uint requestIndex);
    event WorkApprovalSubmitted(uint requestIndex, bool approval);
    event PriceChangeProposed(uint priceProposalIndex);
    event PriceChangeAccepted(uint priceProposalIndex, uint price);

    constructor(
        DAO dao,
        Proposals proposalsContract_,
        uint price_
    ) Availability(dao) {
        price = price_;
        proposalsContract = proposalsContract_;
    }

    /// Accept work request with fee
    function requestWork(string calldata requestPostId) external payable {
        require(msg.value >= price, "Insufficient fee");
        uint requestIndex = requestCount++;
        WorkRequest storage request = requests[requestIndex];
        request.customer = msg.sender;
        request.fee = msg.value;
        request.stakeIndex = assignWork();
        request.requestPostId = requestPostId;
        emit WorkAssigned(requestIndex, request.stakeIndex);
    }

    /// Accept work evidence from worker
    function submitWorkEvidence(
        uint requestIndex,
        string calldata evidencePostId,
        Citation[] calldata citations
    ) external {
        WorkRequest storage request = requests[requestIndex];
        require(
            request.status == WorkStatus.Requested,
            "Status must be Requested"
        );
        AvailabilityStake storage stake = stakes[request.stakeIndex];
        require(
            stake.worker == msg.sender,
            "Worker can only submit evidence for work they are assigned"
        );
        request.status = WorkStatus.EvidenceSubmitted;
        request.evidencePostId = evidencePostId;
        for (uint i = 0; i < citations.length; i++) {
            request.citations.push(citations[i]);
        }
        emit WorkEvidenceSubmitted(requestIndex);
    }

    /// Accept work approval/disapproval from customer
    function submitWorkApproval(
        uint requestIndex,
        bool approval
    ) external virtual {
        WorkRequest storage request = requests[requestIndex];
        require(
            request.status == WorkStatus.EvidenceSubmitted,
            "Status must be EvidenceSubmitted"
        );
        AvailabilityStake storage stake = stakes[request.stakeIndex];
        request.status = WorkStatus.ApprovalSubmitted;
        request.approval = approval;
        // Make work evidence post
        Author[] memory authors = new Author[](1);
        authors[0] = Author(1000000, stake.worker);
        dao.addPost(authors, request.evidencePostId, request.citations);
        emit WorkApprovalSubmitted(requestIndex, approval);
        // Initiate validation pool
        uint poolIndex = dao.initiateValidationPool{value: request.fee}(
            request.evidencePostId,
            POOL_DURATION,
            [uint256(1), uint256(3)],
            [uint256(1), uint256(2)],
            100,
            true,
            false,
            ""
        );
        // We have an approval from stake.worker to transfer up to stake.amount
        dao.delegatedStakeOnValidationPool(
            poolIndex,
            stake.worker,
            stake.amount,
            true
        );
    }

    /// Initiate a new proposal to change the price for this work contract.
    /// This takes a postId; DAO.addPost should be called before or concurrently with this.
    function proposeNewPrice(
        uint newPrice,
        string calldata postId,
        uint[3] calldata durations
    ) external payable {
        uint priceProposalIndex = priceProposalCount++;
        PriceProposal storage priceProposal = priceProposals[
            priceProposalIndex
        ];
        priceProposal.price = newPrice;
        priceProposal.proposalIndex = proposalsContract.propose{
            value: msg.value
        }(postId, durations, true, abi.encode(priceProposalIndex));
        emit PriceChangeProposed(priceProposalIndex);
    }

    function onProposalAccepted(
        uint, // stakedFor
        uint, // stakedAgainst
        bytes calldata callbackData
    ) external returns (uint) {
        uint priceProposalIndex = abi.decode(callbackData, (uint));
        PriceProposal storage priceProposal = priceProposals[
            priceProposalIndex
        ];
        price = priceProposal.price;
        emit PriceChangeAccepted(priceProposalIndex, price);
        return 0;
    }
}
