// SPDX-License-Identifier: Unlicense
pragma solidity ^0.8.24;

import {ERC20} from "@openzeppelin/contracts/token/ERC20/ERC20.sol";
import "./Reputation.sol";
import "./Bench.sol";
import "./Forum.sol";
import "../interfaces/IAcceptAvailability.sol";
import "../interfaces/IOnValidate.sol";

import "hardhat/console.sol";

contract DAO {
    Reputation rep;
    Forum forum;
    Bench bench;
    mapping(uint => address) public members;
    uint public memberCount;
    mapping(address => bool) public isMember;

    event ValidationPoolInitiated(uint poolIndex);
    event ValidationPoolResolved(
        uint poolIndex,
        bool votePasses,
        bool quorumMet
    );
    event PostAdded(string id);

    constructor(Reputation reputation_, Forum forum_, Bench bench_) {
        rep = reputation_;
        forum = forum_;
        bench = bench_;
        rep.registerDAO(this);
        forum.registerDAO(this);
        bench.registerDAO(this);
    }

    function emitValidationPoolInitiated(uint poolIndex) public {
        emit ValidationPoolInitiated(poolIndex);
    }

    function emitValidationPoolResolved(
        uint poolIndex,
        bool votePasses,
        bool quorumMet
    ) public {
        emit ValidationPoolResolved(poolIndex, votePasses, quorumMet);
    }

    function emitPostAdded(string memory id) public {
        emit PostAdded(id);
    }

    function update(address from, address to, uint256 value) public {
        require(
            msg.sender == address(forum) || msg.sender == address(bench),
            "Only DAO core contracts may call update"
        );
        rep.update(from, to, value);
    }

    function mint(address account, uint256 value) public {
        require(
            msg.sender == address(forum) || msg.sender == address(bench),
            "Only DAO core contracts may call mint"
        );
        rep.mint(account, value);
    }

    function burn(address account, uint256 value) public {
        require(
            msg.sender == address(forum) || msg.sender == address(bench),
            "Only DAO core contracts may call burn"
        );
        rep.burn(account, value);
    }

    function registerMember(address account) public {
        require(
            msg.sender == address(forum) || msg.sender == address(bench),
            "Only DAO core contracts may call registerMember"
        );
        if (!isMember[account]) {
            members[memberCount++] = account;
            isMember[account] = true;
        }
    }

    function balanceOf(address account) public view returns (uint256) {
        return rep.balanceOf(account);
    }

    function totalSupply() public view returns (uint256) {
        return rep.totalSupply();
    }

    function allowance(
        address owner,
        address spender
    ) public view returns (uint256) {
        return rep.allowance(owner, spender);
    }

    function forwardAllowance(
        address owner,
        address to,
        uint256 amount
    ) public {
        rep.spendAllowance(owner, msg.sender, amount);
        rep.approve(owner, to, rep.allowance(owner, to) + amount);
    }

    /// Authorize a contract to transfer REP, and call that contract's acceptAvailability method
    function stakeAvailability(
        address to,
        uint256 value,
        uint duration
    ) external returns (bool) {
        uint refund = IAcceptAvailability(to).acceptAvailability(
            msg.sender,
            value,
            duration
        );
        rep.approve(
            msg.sender,
            to,
            rep.allowance(msg.sender, to) + value - refund
        );
        return true;
    }

    function propagateReputation(string memory postId, int amount) public {
        forum.propagateReputation(postId, amount, false, 0);
    }

    function distributeFeeAmongMembers() public payable {
        uint allocated;
        for (uint i = 0; i < memberCount; i++) {
            address member = members[i];
            uint share;
            if (i < memberCount - 1) {
                share = (msg.value * balanceOf(member)) / totalSupply();
                allocated += share;
            } else {
                // Due to rounding, give the remainder to the last member
                share = msg.value - allocated;
            }
            // TODO: For efficiency this could be modified to hold the funds for recipients to withdraw
            payable(member).transfer(share);
        }
    }

    function validationPools(
        uint poolIndex
    )
        public
        view
        returns (
            uint id,
            address sender,
            uint stakeCount,
            ValidationPoolParams memory params,
            ValidationPoolProps memory props,
            bool callbackOnValidate,
            bytes memory callbackData
        )
    {
        return bench.validationPools(poolIndex);
    }

    function validationPoolCount() public view returns (uint) {
        return bench.validationPoolCount();
    }

    function initiateValidationPool(
        string calldata postId,
        uint duration,
        uint[2] calldata quorum, // [Numerator, Denominator]
        uint[2] calldata winRatio, // [Numerator, Denominator]
        uint bindingPercent,
        bool redistributeLosingStakes,
        bool callbackOnValidate,
        bytes calldata callbackData
    ) external payable returns (uint) {
        return
            bench.initiateValidationPool{value: msg.value}(
                msg.sender,
                postId,
                duration,
                quorum,
                winRatio,
                bindingPercent,
                redistributeLosingStakes,
                callbackOnValidate,
                callbackData
            );
    }

    function stakeOnValidationPool(
        uint poolIndex,
        uint256 amount,
        bool inFavor
    ) public {
        require(
            balanceOf(msg.sender) >= amount,
            "Insufficient REP balance to cover stake"
        );
        // TODO: Encumber tokens
        bench.stakeOnValidationPool(poolIndex, msg.sender, amount, inFavor);
    }

    /// Accept reputation stakes toward a validation pool
    function delegatedStakeOnValidationPool(
        uint poolIndex,
        address owner,
        uint256 amount,
        bool inFavor
    ) public {
        if (allowance(owner, msg.sender) < amount) {
            amount = allowance(owner, msg.sender);
        }
        rep.spendAllowance(owner, msg.sender, amount);
        bench.stakeOnValidationPool(poolIndex, owner, amount, inFavor);
    }

    function evaluateOutcome(uint poolIndex) public returns (bool) {
        return bench.evaluateOutcome(poolIndex);
    }

    function onValidate(
        address target,
        bool votePasses,
        bool quorumMet,
        uint stakedFor,
        uint stakedAgainst,
        bytes calldata callbackData
    ) public {
        require(
            msg.sender == address(forum) || msg.sender == address(bench),
            "Only DAO core contracts may call onValidate"
        );
        IOnValidate(target).onValidate(
            votePasses,
            quorumMet,
            stakedFor,
            stakedAgainst,
            callbackData
        );
    }

    function addPost(
        Author[] calldata authors,
        string calldata postId,
        Citation[] calldata citations
    ) public {
        forum.addPost(msg.sender, authors, postId, citations);
    }

    function posts(
        string calldata postId
    ) public view returns (string memory id, address sender, uint reputation) {
        return forum.posts(postId);
    }

    function postCount() public view returns (uint) {
        return forum.postCount();
    }

    function postIds(uint postIndex) public view returns (string memory) {
        return forum.postIds(postIndex);
    }

    function getPostAuthors(
        string calldata postId
    ) public view returns (Author[] memory) {
        return forum.getPostAuthors(postId);
    }
}

/// Convenience contract to extend for other contracts that will be initialized to
/// interact with a DAO contract.
contract DAOContract {
    DAO immutable dao;

    constructor(DAO dao_) {
        dao = dao_;
    }
}
