const {
  time,
  loadFixture,
} = require('@nomicfoundation/hardhat-toolbox/network-helpers');
const { expect } = require('chai');
const { ethers } = require('hardhat');
const deployDAO = require('./util/deploy-dao');

describe('Work1', () => {
  const WORK1_PRICE = 100;
  const STAKE_DURATION = 60;
  async function deploy() {
    // Contracts are deployed using the first signer/account by default
    const [account1, account2] = await ethers.getSigners();

    const { dao } = await deployDAO();
    const Proposals = await ethers.getContractFactory('Proposals');
    const proposals = await Proposals.deploy(dao.target);
    const Work1 = await ethers.getContractFactory('Work1');
    const work1 = await Work1.deploy(dao.target, proposals.target, WORK1_PRICE);

    await dao.addPost([{ weightPPM: 1000000, authorAddress: account1 }], 'some-content-id', []);
    const callbackData = ethers.AbiCoder.defaultAbiCoder().encode([], []);
    await dao.initiateValidationPool(
      'some-content-id',
      60,
      [1, 3],
      [1, 2],
      100,
      true,
      false,
      callbackData,
      { value: 100 },
    );
    await time.increase(61);
    await dao.evaluateOutcome(0);

    return {
      dao, work1, proposals, account1, account2,
    };
  }

  it('Should deploy', async () => {
    const { dao, work1, account1 } = await loadFixture(deploy);
    expect(dao).to.exist;
    expect(work1).to.exist;
    expect(await dao.memberCount()).to.equal(1);
    expect(await dao.balanceOf(account1)).to.equal(100);
    expect(await dao.totalSupply()).to.equal(100);
    expect(await work1.stakeCount()).to.equal(0);
  });

  describe('Stake availability', () => {
    let dao;
    let work1;
    let account1;

    beforeEach(async () => {
      ({ dao, work1, account1 } = await loadFixture(deploy));
      await expect(dao.stakeAvailability(work1.target, 50, STAKE_DURATION)).to.emit(work1, 'AvailabilityStaked').withArgs(0);
    });

    it('Should be able to stake availability', async () => {
      expect(await dao.balanceOf(account1)).to.equal(100);
      expect(await dao.balanceOf(work1.target)).to.equal(0);
      expect(await dao.allowance(account1, work1.target)).to.equal(50);
      expect(await work1.stakeCount()).to.equal(1);
      const stake = await work1.stakes(0);
      expect(stake.worker).to.equal(account1);
      expect(stake.amount).to.equal(50);
      expect(stake.endTime).to.equal(await time.latest() + STAKE_DURATION);
      expect(await dao.allowance(account1, work1.target)).to.equal(50);
    });

    it('should not be able to stake availability without reputation value', async () => {
      await expect(dao.stakeAvailability(work1.target, 0, STAKE_DURATION)).to.be.revertedWith('No stake provided');
    });

    it('should not be able to call acceptAvailability directly', async () => {
      await expect(work1.acceptAvailability(account1, 50, STAKE_DURATION)).to.be.revertedWith('acceptAvailability must only be called by DAO contract');
    });

    it('should be able to extend the duration of an availability stake before it expires', async () => {
      await time.increase(STAKE_DURATION / 2);
      expect(await work1.stakeCount()).to.equal(1);
      await expect(dao.stakeAvailability(work1.target, 50, STAKE_DURATION)).to.emit(work1, 'AvailabilityStaked').withArgs(0);
      expect(await work1.stakeCount()).to.equal(1);
      expect(await dao.allowance(account1, work1.target)).to.equal(50);
    });

    it('should be able to extend the duration of an availability stake after it expires', async () => {
      await time.increase(STAKE_DURATION * 2);
      expect(await work1.stakeCount()).to.equal(1);
      await dao.stakeAvailability(work1.target, 50, STAKE_DURATION);
      expect(await work1.stakeCount()).to.equal(1);
      expect(await dao.allowance(account1, work1.target)).to.equal(50);
    });

    it('extending a stake before expiration should reset the end time to the new duration from the present', async () => {
      await time.increase(STAKE_DURATION / 2);
      expect(await work1.stakeCount()).to.equal(1);
      await dao.stakeAvailability(work1.target, 50, STAKE_DURATION * 2);
      expect(await work1.stakeCount()).to.equal(1);
      const expectedEndTime = await time.latest() + 2 * STAKE_DURATION;
      const stake = await work1.stakes(0);
      expect(stake.endTime).to.be.within(expectedEndTime - 1, expectedEndTime);
      expect(await dao.allowance(account1, work1.target)).to.equal(50);
    });

    it('extending a stake after expiration should restart the stake for the given duration', async () => {
      await time.increase(STAKE_DURATION * 2);
      expect(await work1.stakeCount()).to.equal(1);
      await dao.stakeAvailability(work1.target, 50, STAKE_DURATION * 2);
      expect(await work1.stakeCount()).to.equal(1);
      const expectedEndTime = await time.latest() + STAKE_DURATION * 2;
      const stake = await work1.stakes(0);
      expect(stake.endTime).to.be.within(expectedEndTime - 1, expectedEndTime);
      expect(await dao.allowance(account1, work1.target)).to.equal(50);
    });
  });

  describe('Request and assign work', () => {
    it('should be able to request work and assign to a worker', async () => {
      const {
        dao, work1, account2,
      } = await loadFixture(deploy);
      await dao.stakeAvailability(work1.target, 50, STAKE_DURATION);
      const requestWork = () => work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await expect(requestWork()).to.emit(work1, 'WorkAssigned').withArgs(0, 0);
      expect(await work1.requestCount()).to.equal(1);
      const request = await work1.requests(0);
      expect(request.customer).to.equal(account2);
      expect(request.requestPostId).to.equal('req-content-id');
    });

    it('should not be able to request work if there are no availability stakes', async () => {
      const {
        work1, account2,
      } = await loadFixture(deploy);
      const requestWork = () => work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await expect(requestWork()).to.be.revertedWith('No available worker stakes');
    });

    it('should not be able to request work if fee is insufficient', async () => {
      const {
        work1, account2,
      } = await loadFixture(deploy);
      const requestWork = () => work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE / 2 });
      await expect(requestWork()).to.be.revertedWith('Insufficient fee');
    });

    it('should not assign work to an expired availability stake', async () => {
      const {
        dao, work1, account2,
      } = await loadFixture(deploy);
      await dao.stakeAvailability(work1.target, 50, STAKE_DURATION);
      const requestWork = () => work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await time.increase(STAKE_DURATION + 1);
      await expect(requestWork()).to.be.revertedWith('No available worker stakes');
    });

    it('should not assign work to the same availability stake twice', async () => {
      const {
        dao, work1, account2,
      } = await loadFixture(deploy);
      await dao.stakeAvailability(work1.target, 50, STAKE_DURATION);
      const requestWork = () => work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await expect(requestWork()).to.emit(work1, 'WorkAssigned').withArgs(0, 0);
      await expect(requestWork()).to.be.revertedWith('No available worker stakes');
    });

    it('after a stake has been assigned work, staking again should create a new stake', async () => {
      const {
        dao, work1, account1, account2,
      } = await loadFixture(deploy);
      await dao.stakeAvailability(work1.target, 50, STAKE_DURATION);
      await work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await time.increase(STAKE_DURATION * 2);
      expect(await work1.stakeCount()).to.equal(1);
      await dao.stakeAvailability(work1.target, 50, STAKE_DURATION);
      expect(await work1.stakeCount()).to.equal(2);
      expect(await dao.allowance(account1, work1.target)).to.equal(100);
    });
  });

  describe('Work evidence and approval/disapproval', () => {
    let dao;
    let work1;
    let account1;
    let account2;

    beforeEach(async () => {
      ({
        dao, work1, account1, account2,
      } = await loadFixture(deploy));
      await dao.stakeAvailability(work1.target, 50, STAKE_DURATION);
    });

    it('should be able to submit work evidence', async () => {
      await work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await expect(work1.submitWorkEvidence(0, 'evidence-content-id', [])).to.emit(work1, 'WorkEvidenceSubmitted').withArgs(0);
    });

    it('should not be able to submit work evidence twice', async () => {
      await work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await expect(work1.submitWorkEvidence(0, 'evidence-content-id', [])).to.emit(work1, 'WorkEvidenceSubmitted').withArgs(0);
      await expect(work1.submitWorkEvidence(0, 'evidence-content-id', [])).to.be.revertedWith('Status must be Requested');
    });

    it('should not be able to submit work evidence for a different worker', async () => {
      await work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await expect(work1.connect(account2).submitWorkEvidence(0, 'evidence-content-id', []))
        .to.be.revertedWith('Worker can only submit evidence for work they are assigned');
    });

    it('should be able to submit work approval', async () => {
      await work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await work1.submitWorkEvidence(0, 'evidence-content-id', []);
      expect(await dao.balanceOf(account1)).to.equal(100);
      expect(await dao.balanceOf(work1.target)).to.equal(0);
      await expect(work1.submitWorkApproval(0, true))
        .to.emit(dao, 'ValidationPoolInitiated').withArgs(1)
        .to.emit(work1, 'WorkApprovalSubmitted').withArgs(0, true);
      expect(await dao.balanceOf(work1.target)).to.equal(0);
      expect(await dao.balanceOf(account1)).to.equal(100);
      const post = await dao.posts('evidence-content-id');
      expect(post.sender).to.equal(work1.target);
      expect(post.id).to.equal('evidence-content-id');
      const postAuthors = await dao.getPostAuthors('evidence-content-id');
      expect(postAuthors).to.have.length(1);
      expect(postAuthors[0].weightPPM).to.equal(1000000);
      expect(postAuthors[0].authorAddress).to.equal(account1);
      const pool = await dao.validationPools(1);
      expect(pool.props.fee).to.equal(WORK1_PRICE);
      expect(pool.sender).to.equal(work1.target);
      expect(pool.props.postId).to.equal('evidence-content-id');
      expect(pool.stakeCount).to.equal(1);
      await time.increase(86401);
      await expect(dao.evaluateOutcome(1)).to.emit(dao, 'ValidationPoolResolved').withArgs(1, true, true);
      expect(await dao.balanceOf(account1)).to.equal(200);
    });

    it('should be able to submit work disapproval', async () => {
      await work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await work1.submitWorkEvidence(0, 'evidence-content-id', []);
      await expect(work1.submitWorkApproval(0, false))
        .to.emit(dao, 'ValidationPoolInitiated').withArgs(1)
        .to.emit(work1, 'WorkApprovalSubmitted').withArgs(0, false);
    });

    it('should not be able to submit work approval/disapproval twice', async () => {
      await work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await work1.submitWorkEvidence(0, 'evidence-content-id', []);
      await expect(work1.submitWorkApproval(0, true)).to.emit(dao, 'ValidationPoolInitiated').withArgs(1);
      await expect(work1.submitWorkApproval(0, true)).to.be.revertedWith('Status must be EvidenceSubmitted');
    });

    it('should not be able to submit work evidence after work approval', async () => {
      await work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await work1.submitWorkEvidence(0, 'evidence-content-id', []);
      await expect(work1.submitWorkApproval(0, true)).to.emit(dao, 'ValidationPoolInitiated').withArgs(1);
      await expect(work1.submitWorkEvidence(0, 'evidence-content-id', [])).to.be.revertedWith('Status must be Requested');
    });

    it('should not be able to submit work approval/disapproval before work evidence', async () => {
      await work1.connect(account2).requestWork('req-content-id', { value: WORK1_PRICE });
      await expect(work1.submitWorkApproval(0, true)).to.be.revertedWith('Status must be EvidenceSubmitted');
    });
  });

  describe('Propose new price', () => {
    it('can propose a new price', async () => {
      const {
        proposals, work1,
      } = await loadFixture(deploy);
      expect(await proposals.proposalCount()).to.equal(0);
      await work1.proposeNewPrice(12345, 'content-id', [1, 1, 1]);
      expect(await proposals.proposalCount()).to.equal(1);
    });

    // TODO: can accept a new price
  });
});
