import {
  useCallback, useState,
} from 'react'; import Button from 'react-bootstrap/Button';
import { Web3 } from 'web3';
import Tab from 'react-bootstrap/Tab';
import Tabs from 'react-bootstrap/Tabs';
import Stack from 'react-bootstrap/Stack';

import WorkContract from '../work-contracts/WorkContract';
import Proposals from '../Proposals';
import ImportPaper from '../ImportPaper';
import ImportPapersByAuthor from '../ImportPapersByAuthor';
import getAddressName from '../../utils/get-address-name';
import ImportMatrixEvent from '../ImportMatrixEvent';
import AddPostModal from '../posts/AddPostModal';
import ViewPostModal from '../posts/ViewPostModal';
import Post from '../../utils/Post';
import useMainContext from '../../contexts/useMainContext';

function MainTabs() {
  const {
    provider, chainId, account, members, reputation, setReputation, posts, validationPools,
    DAORef, work1, work2, onboarding,
  } = useMainContext();

  const [showAddPost, setShowAddPost] = useState(false);
  const [showViewPost, setShowViewPost] = useState(false);
  const [viewPost, setViewPost] = useState({});

  /* -------------------------------------------------------------------------------- */
  /* --------------------------- BEGIN UI ACTIONS ----------------------------------- */
  /* -------------------------------------------------------------------------------- */

  const initiateValidationPool = useCallback(async (postId, poolDuration) => {
    const web3 = new Web3(provider);
    await DAORef.current.methods.initiateValidationPool(
      postId,
      poolDuration ?? 3600,
      [1, 3],
      [1, 2],
      100,
      true,
      false,
      web3.eth.abi.encodeParameter('bytes', '0x00'),
    ).send({
      from: account,
      gas: 1000000,
      value: 10000,
    });
  }, [provider, DAORef, account]);

  const stake = useCallback(async (poolIndex, amount, inFavor) => {
    console.log(`Attempting to stake ${amount} ${inFavor ? 'for' : 'against'} pool ${poolIndex}`);
    await DAORef.current.methods.stakeOnValidationPool(poolIndex, amount, inFavor).send({
      from: account,
      gas: 999999,
    });

    // Since this is the result we expect from the server, we preemptively set it here.
    // We can let this value be negative -- this would just mean we'll be getting
    // at least one error from the server, and a corrected reputation.
    setReputation((current) => current - BigInt(amount));
  }, [DAORef, account, setReputation]);

  const stakeHalfInFavor = useCallback(async (poolIndex) => {
    await stake(poolIndex, reputation / BigInt(2), true);
  }, [stake, reputation]);

  const evaluateOutcome = useCallback(async (poolIndex) => {
    await DAORef.current.methods.evaluateOutcome(poolIndex).send({
      from: account,
      gas: 10000000,
    });
  }, [DAORef, account]);

  const handleShowAddPost = () => setShowAddPost(true);

  const handleShowViewPost = useCallback(async ({ id }) => {
    const post = await Post.read(id);
    setViewPost(post);
    setShowViewPost(true);
  }, [setViewPost, setShowViewPost]);

  /* -------------------------------------------------------------------------------- */
  /* --------------------------- END UI ACTIONS ------------------------------------- */
  /* -------------------------------------------------------------------------------- */

  return (
    <>

      <AddPostModal show={showAddPost} setShow={setShowAddPost} postToBlockchain />

      <ViewPostModal show={showViewPost} setShow={setShowViewPost} post={viewPost} />
      <Tabs>
        <Tab eventKey="admin" title="Admin">
          <h2>Members</h2>
          <div>
            {`Members count: ${members.length}`}
          </div>
          <div>
            <table className="table">
              <thead>
                <tr>
                  <th>ID</th>
                  <th>Reputation</th>
                </tr>
              </thead>
              <tbody>
                {members.filter((x) => !!x).map((member) => (
                  <tr key={member.id}>
                    <td>{member.id}</td>
                    <td>{member.reputation.toString()}</td>
                  </tr>
                ))}
              </tbody>
            </table>
          </div>
          {' '}
          <h2>Posts</h2>
          <div>
            <Button onClick={handleShowAddPost}>Add Post</Button>
          </div>
          <div>
            {`Posts count: ${posts.length}`}
          </div>
          <div>
            <table className="table">
              <thead>
                <tr>
                  <th>ID</th>
                  <th>Authors</th>
                  <th>Sender</th>
                  <th>Reputation</th>
                  <th>Actions</th>
                </tr>
              </thead>
              <tbody>
                {posts.filter((x) => !!x).map((post) => (
                  <tr key={post.id}>
                    <td>{post.id.toString()}</td>
                    <td>
                      <Stack>
                        {post.authors.map(({ authorAddress, weightPPM }) => (
                          <div key={authorAddress}>
                            {getAddressName(chainId, authorAddress)}
                            {' '}
                            {Number(weightPPM) / 10000}
                            %
                          </div>
                        ))}
                      </Stack>
                    </td>
                    <td>{getAddressName(chainId, post.sender)}</td>
                    <td>{post.reputation.toString()}</td>
                    <td>
                      <Button onClick={() => handleShowViewPost(post)}>
                        View Post
                      </Button>
                      {' '}
                      Initiate Validation Pool
                      {' '}
                      <Button onClick={() => initiateValidationPool(post.id, 1)}>
                        1s
                      </Button>
                      {' '}
                      <Button onClick={() => initiateValidationPool(post.id, 20)}>
                        20s
                      </Button>
                      {' '}
                      <Button onClick={() => initiateValidationPool(post.id, 60)}>
                        60s
                      </Button>
                    </td>
                  </tr>
                ))}
              </tbody>
            </table>
          </div>
          <h2>Validation Pools</h2>
          <div>
            {`Validation Pool Count: ${validationPools.length}`}
          </div>
          <div>
            <table className="table">
              <thead>
                <tr>
                  <th>ID</th>
                  <th>Post ID</th>
                  <th>Sender</th>
                  <th>Fee</th>
                  <th>Binding</th>
                  <th>Quorum</th>
                  <th>WinRatio</th>
                  <th>
                    Redistribute
                    <br />
                    Losing Stakes
                  </th>
                  <th>Duration</th>
                  <th>End Time</th>
                  <th>
                    Stake
                    <br />
                    Count
                  </th>
                  <th>Status</th>
                  <th>Actions</th>
                </tr>
              </thead>
              <tbody>
                {validationPools.filter((x) => !!x).map((pool) => (
                  <tr key={pool.id}>
                    <td>{pool.id.toString()}</td>
                    <td>{pool.props.postId}</td>
                    <td>{getAddressName(chainId, pool.sender)}</td>
                    <td>{pool.props.fee.toString()}</td>
                    <td>
                      {pool.params.bindingPercent.toString()}
                      %
                    </td>
                    <td>{`${pool.params.quorum[0].toString()}/${pool.params.quorum[1].toString()}`}</td>
                    <td>{`${pool.params.winRatio[0].toString()}/${pool.params.winRatio[1].toString()}`}</td>
                    <td>{pool.params.redistributeLosingStakes.toString()}</td>
                    <td>{pool.params.duration.toString()}</td>
                    <td>{new Date(Number(pool.props.endTime) * 1000).toLocaleString()}</td>
                    <td>{pool.stakeCount.toString()}</td>
                    <td>{pool.status}</td>
                    <td>
                      {!pool.props.resolved && reputation > 0 && pool.timeRemaining > 0 && (
                      <>
                        <Button onClick={() => stakeHalfInFavor(pool.id)}>
                          Stake 1/2 REP
                        </Button>
                        {' '}
                        <Button onClick={() => stake(pool.id, reputation, true)}>
                          Stake All
                        </Button>
                        {' '}
                      </>
                      )}
                      {!pool.props.resolved && (pool.timeRemaining <= 0 || !reputation) && (
                      <Button onClick={() => evaluateOutcome(pool.id)}>
                        Evaluate Outcome
                      </Button>
                      )}
                    </td>
                  </tr>
                ))}
              </tbody>
            </table>
          </div>
        </Tab>
        <Tab eventKey="worker" title="Worker">
          {work1 && (
            <WorkContract
              workContract={work1}
              title="Work Contract 1"
              verb="Work"
              showProposePriceChange
            />
          )}
          {onboarding && (
            <WorkContract
              workContract={onboarding}
              title="Onboarding"
              verb="Onboarding"
              showRequestWork
              showProposePriceChange
            />
          )}
          {work2 && (
            <WorkContract
              workContract={work2}
              title="Work Contract 2"
              verb="Work"
              showProposePriceChange
            />
          )}
        </Tab>
        <Tab eventKey="customer" title="Customer">
          {work1 && (
            <WorkContract
              workContract={work1}
              showAvailabilityActions={false}
              showAvailabilityAmount={false}
              onlyShowAvailable
              title="Work Contract 1"
              verb="Work"
              showRequestWork
            />
          )}
          {work2 && (
            <WorkContract
              workContract={work2}
              showAvailabilityActions={false}
              showAvailabilityAmount={false}
              onlyShowAvailable
              title="Work Contract 2"
              verb="Work"
              showRequestWork
            />
          )}
        </Tab>
        <Tab eventKey="proposals" title="Proposals">
          <Proposals />
        </Tab>
        <Tab eventKey="import" title="Import">
          <h1>Semantic Scholar</h1>
          <ImportPaper />
          <ImportPapersByAuthor />
          <h1>Matrix</h1>
          <ImportMatrixEvent />
        </Tab>
      </Tabs>
    </>
  );
}

export default MainTabs;
