import { useCallback, useContext, useEffect } from 'react';
import { PropTypes } from 'prop-types';
import Button from 'react-bootstrap/Button';

import Web3Context from '../../contexts/Web3Context';
import WorkContractContext from '../../contexts/WorkContractContext';

const getAvailabilityStatus = (stake) => {
  if (stake.assigned) return 'Assigned';
  if (new Date() < new Date(Number(stake.endTime) * 1000)) return 'Available';
  return 'Expired';
};

function AvailabilityStakes({
  showActions, showAmount, onlyShowAvailable,
}) {
  const {
    DAO, reputation, setReputation, account,
  } = useContext(Web3Context);

  const {
    workContract,
    availabilityStakes,
    dispatchAvailabilityStake,
  } = useContext(WorkContractContext);

  const fetchAvailabilityStake = useCallback(async (stakeIndex) => {
    const s = await workContract.methods.stakes(stakeIndex).call();
    Object.assign(s, {
      id: Number(stakeIndex),
      currentUserIsWorker: () => s.worker.toLowerCase() === account.toString().toLowerCase(),
      timeRemaining: new Date(Number(s.endTime) * 1000) - new Date(),
    });
    dispatchAvailabilityStake({ type: 'update', item: s });
    if (s.timeRemaining > 0) {
      setTimeout(() => {
        s.timeRemaining = 0;
        dispatchAvailabilityStake({ type: 'update', item: s });
      }, s.timeRemaining);
    }
    return s;
  }, [workContract, account, dispatchAvailabilityStake]);

  const fetchAvailabilityStakes = useCallback(async () => {
    const count = await workContract.methods.stakeCount().call();
    const promises = [];
    dispatchAvailabilityStake({ type: 'refresh' });
    for (let i = 0; i < count; i += 1) {
      promises.push(fetchAvailabilityStake(i));
    }
    await Promise.all(promises);
  }, [workContract, fetchAvailabilityStake, dispatchAvailabilityStake]);

  useEffect(() => {
    fetchAvailabilityStakes();

    workContract.events.AvailabilityStaked({ fromBlock: 'latest' }).on('data', (event) => {
      console.log('event: availability staked', event);
      fetchAvailabilityStake(event.returnValues.stakeIndex);
    });

    workContract.events.WorkAssigned({ fromBlock: 'latest' }).on('data', (event) => {
      fetchAvailabilityStake(event.returnValues.stakeIndex);
    });
  }, [workContract, fetchAvailabilityStakes, fetchAvailabilityStake]);

  const stakeAvailability = useCallback(async (duration) => {
    const target = workContract.options.address;
    await DAO.methods.stakeAvailability(target, reputation / BigInt(2), duration).send({
      from: account,
      gas: 999999,
    });
    // Note that as with validation pool stakes, we should keep track locally of our reputation
    setReputation(reputation / BigInt(2));
  }, [DAO, workContract, account, reputation, setReputation]);

  const displayData = availabilityStakes.filter((stake) => {
    if (!stake) return false;
    if (!onlyShowAvailable) return true;
    if (getAvailabilityStatus(stake) === 'Available') return true;
    return false;
  });

  return (
    <>
      {showActions && (
        <div>
          Stake Availability:
          {' '}
          {!reputation && <>No reputation available to stake</>}
          {reputation > 0 && (
          <>
            <Button onClick={() => stakeAvailability(300)}>5 Min.</Button>
            {' '}
            <Button onClick={() => stakeAvailability(7200)}>2 Hr.</Button>
          </>
          )}
        </div>
      )}
      <div>
        Availability Stake Count:
        {' '}
        {displayData?.length}
      </div>
      <div>
        <table className="table">
          <thead>
            <tr>
              <th>ID</th>
              <th>Worker</th>
              {showAmount && <th>Amount</th>}
              <th>End Time</th>
              <th>Status</th>
            </tr>
          </thead>
          <tbody>
            {displayData?.filter((x) => !!x).map((s) => (
              <tr key={s.id}>
                <td>{s.id.toString()}</td>
                <td>{s.worker.toString()}</td>
                {showAmount && <td>{s.amount.toString()}</td>}
                <td>{new Date(Number(s.endTime) * 1000).toLocaleString()}</td>
                <td>{getAvailabilityStatus(s)}</td>
              </tr>
            ))}
          </tbody>
        </table>
      </div>
    </>
  );
}

AvailabilityStakes.propTypes = {
  showActions: PropTypes.bool,
  showAmount: PropTypes.bool,
  onlyShowAvailable: PropTypes.bool,
};

AvailabilityStakes.defaultProps = {
  showActions: true,
  showAmount: true,
  onlyShowAvailable: false,
};

export default AvailabilityStakes;
