import {
  useCallback, useContext, useEffect, useRef, useState,
} from 'react';
import Button from 'react-bootstrap/Button';
import Form from 'react-bootstrap/Form';
import Modal from 'react-bootstrap/Modal';
import PropTypes from 'prop-types';

import Web3 from 'web3';
import Web3Context from '../../contexts/Web3Context';
import Post from '../../utils/Post';
import ProposalsArtifact from '../../../contractArtifacts/Proposals.json';
import { getContractAddressByChainId } from '../../utils/contract-config';
import WorkContractContext from '../../contexts/WorkContractContext';

function ProposePriceChangeModal({
  show, setShow, title,
}) {
  const {
    provider, account, chainId,
  } = useContext(Web3Context);
  const { workContract } = useContext(WorkContractContext);
  const [content, setContent] = useState('');
  const [proposedPrice, setProposedPrice] = useState();
  const proposalsContract = useRef();

  useEffect(() => {
    const web3 = new Web3(provider);
    const ProposalsAddress = getContractAddressByChainId(chainId, 'Proposals');
    const contract = new web3.eth.Contract(ProposalsArtifact.abi, ProposalsAddress);
    proposalsContract.current = contract;
  }, [provider, chainId]);

  const handleClose = () => setShow(false);

  const handleSubmit = useCallback(async () => {
    const post = new Post({
      content,
      authors: [{ weightPPM: 1000000, authorAddress: account }],
      embeddedData: {
        proposedPrice,
        nonce: crypto.randomUUID().replaceAll('-', ''),
      },
    });
    // Include metamask signature
    await post.sign(provider, account);
    // Clear the input and hide the modal
    setContent('');
    setShow(false);
    // Write to API
    await post.write();
    // TODO: Make durations configurable
    await workContract.methods.proposeNewPrice(proposedPrice, post.hash, [300, 300, 300]).send({
      from: account,
      gas: 999999,
      value: 200,
    });
  }, [provider, workContract, account, content, setShow, proposedPrice]);

  return (
    <Modal show={show} onHide={handleClose}>
      <Modal.Header closeButton>
        <Modal.Title>{title}</Modal.Title>
      </Modal.Header>
      <Modal.Body>
        <Form>
          <Form.Group className="mb-3">
            <Form.Label>New Price</Form.Label>
            <Form.Control type="text" onChange={(e) => setProposedPrice(e.target.value)} />
          </Form.Group>
          <Form.Group className="mb-3">
            <Form.Label>Explanation</Form.Label>
            <Form.Control
              as="textarea"
              rows={3}
              onChange={(e) => setContent(e.target.value)}
            />
          </Form.Group>
        </Form>

      </Modal.Body>
      <Modal.Footer>
        <Button variant="secondary" onClick={handleClose}>
          Close
        </Button>
        <Button variant="primary" onClick={handleSubmit}>
          Submit
        </Button>
      </Modal.Footer>
    </Modal>
  );
}

ProposePriceChangeModal.propTypes = {
  show: PropTypes.bool.isRequired,
  setShow: PropTypes.func.isRequired,
  title: PropTypes.string,
};

ProposePriceChangeModal.defaultProps = {
  title: 'Propose Price Change',
};

export default ProposePriceChangeModal;
