import {
  useCallback, useContext, useEffect, useMemo, useState,
} from 'react';
import { PropTypes } from 'prop-types';
import Web3 from 'web3';
import Button from 'react-bootstrap/Button';

import Web3Context from '../../contexts/Web3Context';
import useList from '../../utils/List';
import WorkContractContext from '../../contexts/WorkContractContext';
import AvailabilityStakes from './AvailabilityStakes';
import WorkRequests from './WorkRequests';
import ProposePriceChangeModal from './ProposePriceChangeModal';
import PriceProposals from './PriceProposals';

function WorkContract({
  workContract,
  showAvailabilityActions,
  showAvailabilityAmount,
  onlyShowAvailable,
  title,
  verb,
  showRequestWork,
  showProposePriceChange,
}) {
  const [availabilityStakes, dispatchAvailabilityStake] = useList();
  const [priceWei, setPriceWei] = useState();
  const [priceEth, setPriceEth] = useState();
  const { provider } = useContext(Web3Context);
  const [showPriceChangeModal, setShowPriceChangeModal] = useState(false);

  const fetchPrice = useCallback(async () => {
    const web3 = new Web3(provider);
    const fetchedPrice = await workContract.methods.price().call();
    setPriceWei(fetchedPrice);
    setPriceEth(web3.utils.fromWei(fetchedPrice, 'ether'));
  }, [workContract, provider]);

  useEffect(() => {
    const web3 = new Web3(provider);

    fetchPrice();

    const onPriceChangeAccepted = (event) => {
      setPriceWei(event.returnValues.price);
      setPriceEth(web3.utils.fromWei(event.returnValues.price, 'ether'));
    };

    workContract.events.PriceChangeAccepted({ fromBlock: 'latest' }).on('data', onPriceChangeAccepted);

    return () => {
      workContract.events.PriceChangeAccepted().off(onPriceChangeAccepted);
    };
  }, [workContract, provider, fetchPrice]);

  const workContractProviderValue = useMemo(() => ({
    workContract, availabilityStakes, dispatchAvailabilityStake, priceWei, priceEth,
  }), [workContract, availabilityStakes, dispatchAvailabilityStake, priceWei, priceEth]);

  return (
    <WorkContractContext.Provider value={workContractProviderValue}>
      {showProposePriceChange && (
        <ProposePriceChangeModal
          show={showPriceChangeModal}
          setShow={setShowPriceChangeModal}
        />
      )}
      <h2>{title}</h2>
      <h3>{`Price: ${priceEth} ETH`}</h3>
      {showProposePriceChange && (
        <>
          <Button onClick={() => setShowPriceChangeModal(true)}>
            Propose New Price
          </Button>
          <PriceProposals />
        </>
      )}
      <AvailabilityStakes
        showActions={showAvailabilityActions}
        showAmount={showAvailabilityAmount}
        onlyShowAvailable={onlyShowAvailable}
      />
      <WorkRequests verb={verb} showRequestWork={showRequestWork} />
    </WorkContractContext.Provider>
  );
}

WorkContract.propTypes = {
  workContract: PropTypes.any.isRequired, // eslint-disable-line react/forbid-prop-types
  showRequestWork: PropTypes.bool,
  title: PropTypes.string.isRequired,
  verb: PropTypes.string.isRequired,
  showAvailabilityActions: PropTypes.bool,
  showAvailabilityAmount: PropTypes.bool,
  onlyShowAvailable: PropTypes.bool,
  showProposePriceChange: PropTypes.bool,
};

WorkContract.defaultProps = {
  showRequestWork: false,
  showAvailabilityActions: true,
  showAvailabilityAmount: true,
  onlyShowAvailable: false,
  showProposePriceChange: false,
};

export default WorkContract;
