import {
  useCallback, useEffect, useState, useMemo, useRef,
} from 'react';
import { useSDK } from '@metamask/sdk-react';
import { Web3 } from 'web3';
import PropTypes from 'prop-types';

import Web3Context from './Web3Context';
import useList from '../utils/List';
import { getContractAddressByChainId } from '../utils/contract-config';
import DAOArtifact from '../../contractArtifacts/DAO.json';
import Work1Artifact from '../../contractArtifacts/Work1.json';
import OnboardingArtifact from '../../contractArtifacts/Onboarding.json';
import Work2Artifact from '../../contractArtifacts/Work2.json';

function MainContextProvider({ children }) {
  const {
    sdk, connected, provider, chainId, account, balance,
  } = useSDK();
  const [DAO, setDAO] = useState();
  const [work1, setWork1] = useState();
  const [work2, setWork2] = useState();
  const [onboarding, setOnboarding] = useState();
  const DAORef = useRef();
  const [balanceEther, setBalanceEther] = useState();
  const [reputation, setReputation] = useState();
  const [totalReputation, setTotalReputation] = useState();
  const [posts, dispatchPost] = useList();
  const [validationPools, dispatchValidationPool] = useList();
  const [members, dispatchMember] = useList();
  const web3ProviderValue = useMemo(() => ({
    provider,
    reputation,
    setReputation,
    totalReputation,
    account,
    chainId,
    posts,
    DAO,
    DAORef,
    work1,
    work2,
    onboarding,
    members,
    validationPools,
    balanceEther,
    sdk,
    connected,
  }), [
    provider, reputation, setReputation, account, chainId, posts, balanceEther, totalReputation,
    DAO, DAORef, work1, work2, onboarding, members, validationPools, sdk, connected]);

  useEffect(() => {
    if (!provider || balance === undefined) return;
    const web3 = new Web3(provider);
    setBalanceEther(web3.utils.fromWei(balance, 'ether'));
  }, [provider, balance]);

  /* -------------------------------------------------------------------------------- */
  /* --------------------------- BEGIN FETCHERS ------------------------------------- */
  /* -------------------------------------------------------------------------------- */

  const fetchReputation = useCallback(async () => {
    setReputation(await DAORef.current.methods.balanceOf(account).call());
    setTotalReputation(await DAORef.current.methods.totalSupply().call());
  }, [DAORef, account]);

  const fetchPost = useCallback(async (postId) => {
    const post = await DAORef.current.methods.posts(postId).call();
    post.authors = await DAORef.current.methods.getPostAuthors(postId).call();
    dispatchPost({ type: 'updateById', item: post });
    return post;
  }, [DAORef, dispatchPost]);

  const fetchPostId = useCallback(async (postIndex) => {
    const postId = await DAORef.current.methods.postIds(postIndex).call();
    return postId;
  }, [DAORef]);

  const fetchPosts = useCallback(async () => {
    const count = await DAORef.current.methods.postCount().call();
    let promises = [];
    dispatchPost({ type: 'refresh' });
    for (let i = 0; i < count; i += 1) {
      promises.push(fetchPostId(i));
    }
    const postIds = await Promise.all(promises);
    promises = [];
    postIds.forEach((postId) => {
      promises.push(fetchPost(postId));
    });
  }, [DAORef, dispatchPost, fetchPost, fetchPostId]);

  const fetchValidationPool = useCallback(async (poolIndex) => {
    const getPoolStatus = (pool) => {
      if (pool.props.resolved) {
        return pool.props.outcome ? 'Accepted' : 'Rejected';
      }
      return pool.timeRemaining > 0 ? 'In Progress' : 'Ready to Evaluate';
    };
    const pool = await DAORef.current.methods.validationPools(poolIndex).call();
    pool.id = Number(pool.id);
    pool.timeRemaining = new Date(Number(pool.props.endTime) * 1000) - new Date();
    pool.status = getPoolStatus(pool);
    dispatchValidationPool({ type: 'update', item: pool });

    // When remaing time expires, we want to update the status for this pool
    if (pool.timeRemaining > 0) {
      setTimeout(() => {
        pool.timeRemaining = 0;
        pool.status = getPoolStatus(pool);
        dispatchValidationPool({ type: 'update', item: pool });
      }, pool.timeRemaining);
    }
  }, [DAORef, dispatchValidationPool]);

  const fetchValidationPools = useCallback(async () => {
    // TODO: Pagination
    // TODO: Memoization
    // TODO: Caching
    const count = await DAORef.current.methods.validationPoolCount().call();
    const promises = [];
    dispatchValidationPool({ type: 'refresh' });
    for (let i = 0; i < count; i += 1) {
      promises.push(fetchValidationPool(i));
    }
    await Promise.all(promises);
  }, [DAORef, dispatchValidationPool, fetchValidationPool]);

  const fetchMember = useCallback(async (memberIndex) => {
    const id = await DAORef.current.methods.members(memberIndex).call();
    const member = { id };
    member.reputation = await DAORef.current.methods.balanceOf(id).call();
    dispatchMember({ type: 'updateById', item: member });
    return member;
  }, [DAORef, dispatchMember]);

  const fetchMembers = useCallback(async () => {
    const count = await DAORef.current.methods.memberCount().call();
    const promises = [];
    dispatchMember({ type: 'refresh' });
    for (let i = 0; i < count; i += 1) {
      promises.push(fetchMember(i));
    }
    await Promise.all(promises);
  }, [DAORef, dispatchMember, fetchMember]);

  /* -------------------------------------------------------------------------------- */
  /* --------------------------- END FETCHERS --------------------------------------- */
  /* -------------------------------------------------------------------------------- */

  // In this effect, we initialize everything and add contract event listeners.
  useEffect(() => {
    if (!provider || !chainId || !account || balance === undefined) return () => {};
    const DAOAddress = getContractAddressByChainId(chainId, 'DAO');
    const Work1Address = getContractAddressByChainId(chainId, 'Work1');
    const OnboardingAddress = getContractAddressByChainId(chainId, 'Onboarding');
    const Work2Address = getContractAddressByChainId(chainId, 'Work2');
    const web3 = new Web3(provider);
    DAORef.current = new web3.eth.Contract(DAOArtifact.abi, DAOAddress);
    setDAO(DAORef.current);
    const work1Contract = new web3.eth.Contract(Work1Artifact.abi, Work1Address);
    setWork1(work1Contract);
    const onboardingContract = new web3.eth.Contract(OnboardingArtifact.abi, OnboardingAddress);
    setOnboarding(onboardingContract);
    const work2Contract = new web3.eth.Contract(Work2Artifact.abi, Work2Address);
    setWork2(work2Contract);

    fetchReputation();
    fetchMembers();
    fetchPosts();
    fetchValidationPools();

    // const fetchReputationInterval = setInterval(() => {
    //   // console.log('reputation', reputation);
    //   if (reputation !== undefined) {
    //     clearInterval(fetchReputationInterval);
    //     return;
    //   }
    //   fetchReputation();
    // }, 1000);

    /* -------------------------------------------------------------------------------- */
    /* --------------------------- BEGIN EVENT HANDLERS ------------------------------- */
    /* -------------------------------------------------------------------------------- */

    DAORef.current.events.PostAdded({ fromBlock: 'latest' }).on('data', (event) => {
      console.log('event: post added');
      fetchPost(event.returnValues.id);
    });

    DAORef.current.events.ValidationPoolInitiated({ fromBlock: 'latest' }).on('data', (event) => {
      console.log('event: validation pool initiated');
      fetchValidationPool(event.returnValues.poolIndex);
    });

    DAORef.current.events.ValidationPoolResolved({ fromBlock: 'latest' }).on('data', (event) => {
      console.log('event: validation pool resolved');
      fetchReputation();
      fetchValidationPool(event.returnValues.poolIndex);
      fetchMembers();
    });

    work1Contract.events.AvailabilityStaked({ fromBlock: 'latest' }).on('data', () => {
      fetchReputation();
    });

    onboardingContract.events.AvailabilityStaked({ fromBlock: 'latest' }).on('data', () => {
      fetchReputation();
    });

    return () => {
      DAORef.current.events.PostAdded().off();
      DAORef.current.events.ValidationPoolInitiated().off();
      DAORef.current.events.ValidationPoolResolved().off();
      work1Contract.events.AvailabilityStaked().off();
      onboardingContract.events.AvailabilityStaked().off();
    };
  }, [provider, account, chainId, balance, dispatchValidationPool, dispatchPost,
    DAORef, setDAO, setWork1, setOnboarding, setWork2,
    fetchPost, fetchPosts, fetchReputation, fetchValidationPool, fetchValidationPools, fetchMembers,
  ]);

  /* -------------------------------------------------------------------------------- */
  /* --------------------------- END MAIN INITIALIZION EFFECT ----------------------- */
  /* -------------------------------------------------------------------------------- */
  return (
    <Web3Context.Provider value={web3ProviderValue}>
      {children}
    </Web3Context.Provider>
  );
}

MainContextProvider.propTypes = {
  children: PropTypes.node.isRequired,
};

export default MainContextProvider;
